#!/usr/bin/env python3
##
# This file is part of ArmarX.
#
# ArmarX is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2 as
# published by the Free Software Foundation.
#
# ArmarX is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
# @package    ArmarXCore::scripts
# @author     Mirko Waechter (waechter at kit dot edu)
# @author     Rainer Kartmann (rainer dot kartmann at kit dot edu)
# @date       2016, 2022
# @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
#             GNU General Public License


import fileinput
import json
import logging
import os
import pathlib
import re
import requests
import subprocess
import sys

import typing as ty


Page = str

log = logging.getLogger(os.path.basename(__file__))


class Source:
    def __init__(self, relative_path: str, line_number: int):
        self.relative_path = relative_path
        self.line_number = line_number


PageSourceDict = ty.Dict[Page, Source]


def join_url(*args) -> str:
    return "/".join(args)


def get_module_info(
        project_base_dir: str,
) -> ty.Dict[str, ty.Any]:
    ws_path = os.environ["ARMARX_WORKSPACE"]
    assert project_base_dir.startswith(ws_path), (project_base_dir, ws_path)
    project_module = os.path.relpath(project_base_dir, ws_path)

    cmd = "axii module show".split(" ") + [project_module, "--json"]

    if sys.version_info < (3, 7):
        kwargs = dict(stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    else:
        kwargs = dict(capture_output=True)
    res = subprocess.run(cmd, **kwargs)
    
    data = json.loads(res.stdout)
    return data


def make_page_source_dict(
        project_base_dir: str,
        search_paths: ty.List[str],
        project_name: str
) -> PageSourceDict:
    regexes = {
        "": r"(@|\\)page ([A-Z0-9a-z\-_@]+) ",
        "mainpage": r"(@|\\)mainpage",
        "group__": r"(@|\\)defgroup ([A-Z0-9a-z\-]+) ",
        "classarmarx_1_1": r"(@|\\)class ([A-Z0-9a-z\-]+)",
    }

    included_extensions = ["dox", "h", "hpp", "cpp", "dox__", "dox.in"]

    page_source_dict: PageSourceDict = {}

    for path in search_paths:
        for root, dir_names, filenames in os.walk(path):
            # log.info(root, dir_names, filenames)
            for filename in filenames:
                if not any(filename.endswith(f".{ext}") for ext in included_extensions):
                    # log.info(f"Skipping file with extension '{extension}'.")
                    continue

                rel_file_path = os.path.join(root, filename)
                try:
                    with open(rel_file_path) as file:
                        lines = file.readlines()

                    for line_number, line in enumerate(lines):
                        line_number += 1
                        for regex_key, regex in regexes.items():
                            matches = list(re.finditer(regex, line))

                            for match in matches:
                                project_base_dir = str(pathlib.Path(project_base_dir).resolve())
                                relative_path = os.path.relpath(rel_file_path, project_base_dir)

                                if regex_key == "mainpage":
                                    if project_name == "armarx_documentation":
                                        page = "index"
                                    else:
                                        continue

                                else:
                                    filename = match.group(2)
                                    filename = filename.replace("@ARMARX_PROJECT_NAME@", project_name)
                                    page = regex_key + filename
                                page_source_dict[page] = Source(relative_path=relative_path, line_number=line_number)

                except UnicodeDecodeError as e:
                    # log.info(f"Reading {rel_file_path} failed: {e}")
                    pass

    return page_source_dict


class EditLink:
    def __init__(self, base_url: str, branch: str):
        self.base_url = base_url
        self.branch = branch

    @classmethod
    def from_module_info(cls, module_info: ty.Dict, verbose=0) -> "EditLink":
        base_url = module_info["general"]["url"]

        branch, source = None, None

        if not branch:
            branch, source = cls._branch_from_module_info(module_info=module_info)
        if not branch:
            branch, source = cls._branch_from_gitlab_url(base_url=base_url)
        if not branch:
            branch, source = "master", ""

        if verbose >= 1:
            log.info(f"Use default branch '{branch}'{source}.")
        return cls(base_url=base_url, branch=branch)

    def edit_url_base(self) -> str:
        return join_url(self.base_url, "-", "edit", self.branch)

    @classmethod
    def _branch_from_module_info(cls, module_info):
        try:
            return module_info["update"]["git"]["default_checkout_ref"], " from module definition"
        except KeyError:
            return None, None

    @classmethod
    def _branch_from_gitlab_url(cls, base_url: str):
        https = "https://"
        assert base_url.startswith(https), base_url
        url_split = base_url[len(https):].split("/")

        host_url = url_split[0]
        project_path = "%2F".join(url_split[1:])

        api_base_url = https + join_url(host_url, "api/v4")
        url = join_url(api_base_url, "projects", project_path)

        response = requests.get(url)
        if response.ok:
            data = json.loads(response.content)
            return data["default_branch"], " from GitLab REST API"
        else:
            return None, None


def update_pages(
        page_source_dict: PageSourceDict,
        module_info: ty.Dict,
        build_dir: str,
) -> int:
    doxygen_search_text = '<li class="footer">'

    build_doxygen_path = os.path.join(build_dir, "doxygen/html")

    link = EditLink.from_module_info(module_info)

    change_count = 0

    for page, source in page_source_dict.items():
        changed_file_path = os.path.join(build_doxygen_path, f"{page}.html")

        link_base_url = join_url(link.edit_url_base(), source.relative_path)
        link_url = "{}#L{}".format(link_base_url, source.line_number)

        added_text = f" <A href='{link_url}' target='_blank'><b>Edit Page</b></A> - "
        replace_text = doxygen_search_text + added_text

        if os.path.isfile(changed_file_path):
            # log.info("Changing file", changed_file_path)
            change_count += 1

            with fileinput.FileInput(changed_file_path, inplace=True) as file:
                for line in file:
                    print(line.replace(doxygen_search_text, replace_text), end='')

    return change_count


def main():

    if len(sys.argv) < 5:
        log.error("Required parameters: build_dir project_name project_base_dir search_path [search_path2 ...] ")
        exit(1)

    build_dir: str = sys.argv[1]
    project_name: str = sys.argv[2]
    project_base_dir: str = sys.argv[3]
    search_paths: ty.List[str] = sys.argv[4:]  # ["etc/doxygen", "source"]

    # Fix during update of Axii module.
    if "ArmarXDoc_DIR" in os.environ:
        os.environ["armarx_documentation_DIR"] = os.environ["ArmarXDoc_DIR"]

    project_dir_env_var = os.environ.get(f"{project_name}_DIR", None)
    log.info(f"name: '{project_name}', project_dir_env_var: '{project_dir_env_var}'")
    if project_dir_env_var is not None:
        assert os.path.basename(project_dir_env_var) == "build", (
            project_dir_env_var, os.path.basename(project_dir_env_var))
        new_project_base_dir = os.path.dirname(project_dir_env_var)
        if new_project_base_dir != project_base_dir:
            log.info(f"Using project_base_dir='{new_project_base_dir}' instead of '{project_base_dir}'")
        project_base_dir = new_project_base_dir

    module_info = get_module_info(project_base_dir=project_base_dir)

    page_source_map = make_page_source_dict(
        project_base_dir=project_base_dir,
        search_paths=search_paths,
        project_name=project_name,
    )

    if True:
        log.info(f"Page source map:")
        for page, source in page_source_map.items():
            log.info(f"- Page '{page}': source '{source.relative_path}' #{source.line_number}")

    num_changes = update_pages(
        page_source_dict=page_source_map,
        module_info=module_info,
        build_dir=build_dir,
    )

    log.info(f"Added {num_changes} source page links for project '{project_name}'.")


if __name__ == '__main__':
    logging.basicConfig(level=logging.INFO, format='[%(levelname)s] %(message)s')
    main()
