/*
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @package    ArmarXCore::ArmarXObjects::EmergencyStop
 * @author     Stefan Reither ( stef dot reither at web dot de )
 * @date       2016
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

#include "EmergencyStop.h"

#include <mutex>

namespace armarx
{

    void
    EmergencyStopMaster::setEmergencyStopState(EmergencyStopState state, const ::Ice::Current&)
    {
        const std::unique_lock lock{ss2StateMutex};

        currentSS2State = state;

        if (currentSS2State == EmergencyStopState::eEmergencyStopActive)
        {
            lastSS2Activation = armarx::core::time::DateTime::Now();
        }

        this->emergencyStopTopic->reportEmergencyStopState(currentSS2State);
    }

    EmergencyStopState
    EmergencyStopMaster::trySetEmergencyStopState(EmergencyStopState targetState,
                                                  const ::Ice::Current&)
    {
        armarx::core::time::Duration safeDuration = armarx::core::time::Duration::MilliSeconds(500);

        // If the SS2 should be released (target state = inactive) we check whether the last
        // activation happened at least before safeDuration.
        if (targetState == EmergencyStopState::eEmergencyStopInactive and
            armarx::core::time::DateTime::Now() > lastSS2Activation + safeDuration)
        {
            setEmergencyStopState(targetState);
        }

        return getEmergencyStopState();
    }

    EmergencyStopState
    EmergencyStopMaster::getEmergencyStopState(const ::Ice::Current& /*unused*/) const
    {
        const std::shared_lock lock{ss2StateMutex};

        return currentSS2State;
    }

    void
    EmergencyStopMaster::onInitComponent()
    {
        armarx::ManagedIceObject::offeringTopic(
            getProperty<std::string>("EmergencyStopTopic").getValue());
    }

    void
    EmergencyStopMaster::onConnectComponent()
    {
        emergencyStopTopic = armarx::ManagedIceObject::getTopic<EmergencyStopListenerPrx>(
            getProperty<std::string>("EmergencyStopTopic").getValue());

        // Default state is inactive
        this->setEmergencyStopState(EmergencyStopState::eEmergencyStopInactive);
    }

    armarx::PropertyDefinitionsPtr
    EmergencyStopMaster::createPropertyDefinitions()
    {
        return armarx::PropertyDefinitionsPtr(
            new EmergencyStopPropertyMasterDefinitions(getConfigIdentifier()));
    }

} // namespace armarx
