#include "ice_conversions.h"

#include <ArmarXCore/core/exceptions/local/ExpressionException.h>

namespace armarx::core::eigen
{
    // COL MAJOR

    void
    fromIce(Eigen::Vector2f& e, const Ice::FloatSeq& ice)
    {
        ARMARX_CHECK(ice.size() >= 2);
        e << ice[0], ice[1];
    }

    void
    toIce(const Eigen::Vector2f& e, Ice::FloatSeq& ice)
    {
        ice = std::vector<float>({e(0, 0), e(1, 0)});
    }

    void
    fromIce(Eigen::Vector2d& e, const Ice::DoubleSeq& ice)
    {
        ARMARX_CHECK(ice.size() >= 2);
        e << ice[0], ice[1];
    }

    void
    toIce(const Eigen::Vector2d& e, Ice::DoubleSeq& ice)
    {
        ice = std::vector<double>({e(0, 0), e(1, 0)});
    }

    void
    fromIce(Eigen::Vector3f& e, const Ice::FloatSeq& ice)
    {
        ARMARX_CHECK(ice.size() >= 3);
        e << ice[0], ice[1], ice[2];
    }

    void
    toIce(const Eigen::Vector3f& e, Ice::FloatSeq& ice)
    {
        ice = std::vector<float>({e(0, 0), e(1, 0), e(2, 0)});
    }

    void
    fromIce(Eigen::Vector3d& e, const Ice::DoubleSeq& ice)
    {
        ARMARX_CHECK(ice.size() >= 3);
        e << ice[0], ice[1], ice[2];
    }

    void
    toIce(const Eigen::Vector3d& e, Ice::DoubleSeq& ice)
    {
        ice = std::vector<double>({e(0, 0), e(1, 0), e(2, 0)});
    }

    void
    fromIce(Eigen::Vector4f& e, const Ice::FloatSeq& ice)
    {
        ARMARX_CHECK(ice.size() >= 4);
        e << ice[0], ice[1], ice[2], ice[3];
    }

    void
    toIce(const Eigen::Vector4f& e, Ice::FloatSeq& ice)
    {
        ice = std::vector<float>({e(0, 0), e(1, 0), e(2, 0), e(3, 0)});
    }

    void
    fromIce(Eigen::Vector4d& e, const Ice::DoubleSeq& ice)
    {
        ARMARX_CHECK(ice.size() >= 4);
        e << ice[0], ice[1], ice[2], ice[3];
    }

    void
    toIce(const Eigen::Vector4d& e, Ice::DoubleSeq& ice)
    {
        ice = std::vector<double>({e(0, 0), e(1, 0), e(2, 0), e(3, 0)});
    }

    void
    fromIce(Eigen::Matrix2f& e, const armarx::FloatSeqSeq& ice)
    {
        ARMARX_CHECK(ice.size() >= 2);
        for (const auto& x : ice)
        {
            ARMARX_CHECK(x.size() >= 2);
        }

        e << ice[0][0], ice[0][1], ice[1][0], ice[1][1], ice[2][0], ice[2][1];
    }

    void
    toIce(const Eigen::Matrix2f& e, armarx::FloatSeqSeq& ice)
    {
        ice = std::vector<std::vector<float>>(
            {{e(0, 0), e(0, 1)}, {e(1, 0), e(1, 1)}, {e(2, 0), e(2, 1)}});
    }

    void
    fromIce(Eigen::Matrix2d& e, const armarx::DoubleSeqSeq& ice)
    {
        ARMARX_CHECK(ice.size() >= 2);
        for (const auto& x : ice)
        {
            ARMARX_CHECK(x.size() >= 2);
        }

        e << ice[0][0], ice[0][1], ice[1][0], ice[1][1], ice[2][0], ice[2][1];
    }

    void
    toIce(const Eigen::Matrix2d& e, armarx::DoubleSeqSeq& ice)
    {
        ice = std::vector<std::vector<double>>(
            {{e(0, 0), e(0, 1)}, {e(1, 0), e(1, 1)}, {e(2, 0), e(2, 1)}});
    }

    void
    fromIce(Eigen::Matrix3f& e, const armarx::FloatSeqSeq& ice)
    {
        ARMARX_CHECK(ice.size() >= 3);
        for (const auto& x : ice)
        {
            ARMARX_CHECK(x.size() >= 3);
        }

        e << ice[0][0], ice[0][1], ice[0][2], ice[1][0], ice[1][1], ice[1][2], ice[2][0], ice[2][1],
            ice[2][2];
    }

    void
    toIce(const Eigen::Matrix3f& e, armarx::FloatSeqSeq& ice)
    {
        ice = std::vector<std::vector<float>>({{e(0, 0), e(0, 1), e(0, 2)},
                                               {e(1, 0), e(1, 1), e(1, 2)},
                                               {e(2, 0), e(2, 1), e(2, 2)}});
    }

    void
    fromIce(Eigen::Matrix3d& e, const armarx::DoubleSeqSeq& ice)
    {
        ARMARX_CHECK(ice.size() >= 3);
        for (const auto& x : ice)
        {
            ARMARX_CHECK(x.size() >= 3);
        }

        e << ice[0][0], ice[0][1], ice[0][2], ice[1][0], ice[1][1], ice[1][2], ice[2][0], ice[2][1],
            ice[2][2];
    }

    void
    toIce(const Eigen::Matrix3d& e, armarx::DoubleSeqSeq& ice)
    {
        ice = std::vector<std::vector<double>>({{e(0, 0), e(0, 1), e(0, 2)},
                                                {e(1, 0), e(1, 1), e(1, 2)},
                                                {e(2, 0), e(2, 1), e(2, 2)}});
    }

    void
    fromIce(Eigen::Matrix4f& e, const armarx::FloatSeqSeq& ice)
    {
        ARMARX_CHECK(ice.size() >= 4);
        for (const auto& x : ice)
        {
            ARMARX_CHECK(x.size() >= 4);
        }
        e << ice[0][0], ice[0][1], ice[0][2], ice[0][3], ice[1][0], ice[1][1], ice[1][2], ice[1][3],
            ice[2][0], ice[2][1], ice[2][2], ice[2][3], ice[3][0], ice[3][1], ice[3][2], ice[3][3];
    }

    void
    toIce(const Eigen::Matrix4f& e, armarx::FloatSeqSeq& ice)
    {
        ice = std::vector<std::vector<float>>({{e(0, 0), e(0, 1), e(0, 2), e(0, 3)},
                                               {e(1, 0), e(1, 1), e(1, 2), e(1, 3)},
                                               {e(2, 0), e(2, 1), e(2, 2), e(2, 3)},
                                               {e(3, 0), e(3, 1), e(3, 2), e(3, 3)}});
    }

    void
    fromIce(Eigen::Matrix4d& e, const armarx::DoubleSeqSeq& ice)
    {
        ARMARX_CHECK(ice.size() >= 4);
        for (const auto& x : ice)
        {
            ARMARX_CHECK(x.size() >= 4);
        }
        e << ice[0][0], ice[0][1], ice[0][2], ice[0][3], ice[1][0], ice[1][1], ice[1][2], ice[1][3],
            ice[2][0], ice[2][1], ice[2][2], ice[2][3], ice[3][0], ice[3][1], ice[3][2], ice[3][3];
    }

    void
    toIce(const Eigen::Matrix4d& e, armarx::DoubleSeqSeq& ice)
    {
        ice = std::vector<std::vector<double>>({{e(0, 0), e(0, 1), e(0, 2), e(0, 3)},
                                                {e(1, 0), e(1, 1), e(1, 2), e(1, 3)},
                                                {e(2, 0), e(2, 1), e(2, 2), e(2, 3)},
                                                {e(3, 0), e(3, 1), e(3, 2), e(3, 3)}});
    }

    // ROW MAJOR

    void
    fromIce(Eigen::Matrix<float, 2, 2, Eigen::RowMajor>& e, const armarx::FloatSeqSeq& ice)
    {
        ARMARX_CHECK(ice.size() >= 2);
        for (const auto& x : ice)
        {
            ARMARX_CHECK(x.size() >= 2);
        }

        e << ice[0][0], ice[0][1], ice[1][0], ice[1][1], ice[2][0], ice[2][1];
    }

    void
    toIce(const Eigen::Matrix<float, 2, 2, Eigen::RowMajor>& e, armarx::FloatSeqSeq& ice)
    {
        ice = std::vector<std::vector<float>>(
            {{e(0, 0), e(0, 1)}, {e(1, 0), e(1, 1)}, {e(2, 0), e(2, 1)}});
    }

    void
    fromIce(Eigen::Matrix<double, 2, 2, Eigen::RowMajor>& e, const armarx::DoubleSeqSeq& ice)
    {
        ARMARX_CHECK(ice.size() >= 2);
        for (const auto& x : ice)
        {
            ARMARX_CHECK(x.size() >= 2);
        }

        e << ice[0][0], ice[0][1], ice[1][0], ice[1][1], ice[2][0], ice[2][1];
    }

    void
    toIce(const Eigen::Matrix<double, 2, 2, Eigen::RowMajor>& e, armarx::DoubleSeqSeq& ice)
    {
        ice = std::vector<std::vector<double>>(
            {{e(0, 0), e(0, 1)}, {e(1, 0), e(1, 1)}, {e(2, 0), e(2, 1)}});
    }

    void
    fromIce(Eigen::Matrix<float, 3, 3, Eigen::RowMajor>& e, const armarx::FloatSeqSeq& ice)
    {
        ARMARX_CHECK(ice.size() >= 3);
        for (const auto& x : ice)
        {
            ARMARX_CHECK(x.size() >= 3);
        }

        e << ice[0][0], ice[0][1], ice[0][2], ice[1][0], ice[1][1], ice[1][2], ice[2][0], ice[2][1],
            ice[2][2];
    }

    void
    toIce(const Eigen::Matrix<float, 3, 3, Eigen::RowMajor>& e, armarx::FloatSeqSeq& ice)
    {
        ice = std::vector<std::vector<float>>({{e(0, 0), e(0, 1), e(0, 2)},
                                               {e(1, 0), e(1, 1), e(1, 2)},
                                               {e(2, 0), e(2, 1), e(2, 2)}});
    }

    void
    fromIce(Eigen::Matrix<double, 3, 3, Eigen::RowMajor>& e, const armarx::DoubleSeqSeq& ice)
    {
        ARMARX_CHECK(ice.size() >= 3);
        for (const auto& x : ice)
        {
            ARMARX_CHECK(x.size() >= 3);
        }

        e << ice[0][0], ice[0][1], ice[0][2], ice[1][0], ice[1][1], ice[1][2], ice[2][0], ice[2][1],
            ice[2][2];
    }

    void
    toIce(const Eigen::Matrix<double, 3, 3, Eigen::RowMajor>& e, armarx::DoubleSeqSeq& ice)
    {
        ice = std::vector<std::vector<double>>({{e(0, 0), e(0, 1), e(0, 2)},
                                                {e(1, 0), e(1, 1), e(1, 2)},
                                                {e(2, 0), e(2, 1), e(2, 2)}});
    }

    void
    fromIce(Eigen::Matrix<float, 4, 4, Eigen::RowMajor>& e, const armarx::FloatSeqSeq& ice)
    {
        ARMARX_CHECK(ice.size() >= 4);
        for (const auto& x : ice)
        {
            ARMARX_CHECK(x.size() >= 4);
        }
        e << ice[0][0], ice[0][1], ice[0][2], ice[0][3], ice[1][0], ice[1][1], ice[1][2], ice[1][3],
            ice[2][0], ice[2][1], ice[2][2], ice[2][3], ice[3][0], ice[3][1], ice[3][2], ice[3][3];
    }

    void
    toIce(const Eigen::Matrix<float, 4, 4, Eigen::RowMajor>& e, armarx::FloatSeqSeq& ice)
    {
        ice = std::vector<std::vector<float>>({{e(0, 0), e(0, 1), e(0, 2), e(0, 3)},
                                               {e(1, 0), e(1, 1), e(1, 2), e(1, 3)},
                                               {e(2, 0), e(2, 1), e(2, 2), e(2, 3)},
                                               {e(3, 0), e(3, 1), e(3, 2), e(3, 3)}});
    }

    void
    fromIce(Eigen::Matrix<double, 4, 4, Eigen::RowMajor>& e, const armarx::DoubleSeqSeq& ice)
    {
        ARMARX_CHECK(ice.size() >= 4);
        for (const auto& x : ice)
        {
            ARMARX_CHECK(x.size() >= 4);
        }
        e << ice[0][0], ice[0][1], ice[0][2], ice[0][3], ice[1][0], ice[1][1], ice[1][2], ice[1][3],
            ice[2][0], ice[2][1], ice[2][2], ice[2][3], ice[3][0], ice[3][1], ice[3][2], ice[3][3];
    }

    void
    toIce(const Eigen::Matrix<double, 4, 4, Eigen::RowMajor>& e, armarx::DoubleSeqSeq& ice)
    {
        ice = std::vector<std::vector<double>>({{e(0, 0), e(0, 1), e(0, 2), e(0, 3)},
                                                {e(1, 0), e(1, 1), e(1, 2), e(1, 3)},
                                                {e(2, 0), e(2, 1), e(2, 2), e(2, 3)},
                                                {e(3, 0), e(3, 1), e(3, 2), e(3, 3)}});
    }


} // namespace armarx::core::eigen
