/*
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @package    ArmarXCore::core::Test
 * @author     Raphael Grimm ( raphael dot grimm at kit dot edu )
 * @date       2017
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

#define BOOST_TEST_MODULE ArmarX::Core::ExpressionException
#define ARMARX_BOOST_TEST

#include <ArmarXCore/Test.h>

#include <optional>

#include <SimoxUtility/algorithm/string/string_tools.h>

#include <ArmarXCore/core/exceptions/local/ExpressionException.h>
#include <ArmarXCore/core/test/IceTestHelper.h>

#define check_do_throw(...)                                                                        \
    BOOST_CHECK_THROW(                                                                             \
        do { __VA_ARGS__; } while (0), ::armarx::LocalException)
#define check_no_throw(...) BOOST_CHECK_NO_THROW(do { __VA_ARGS__; } while (0))

BOOST_AUTO_TEST_CASE(test_check_expression)
{
    check_no_throw(ARMARX_CHECK_EXPRESSION(true));
    check_do_throw(ARMARX_CHECK_EXPRESSION(false));
}

BOOST_AUTO_TEST_CASE(test_check)
{
    check_no_throw(ARMARX_CHECK(true));
    check_do_throw(ARMARX_CHECK(false));
}

BOOST_AUTO_TEST_CASE(test_check_less)
{
    check_do_throw(ARMARX_CHECK_LESS(1, 1));
    check_do_throw(ARMARX_CHECK_LESS(1, -1));
    check_do_throw(ARMARX_CHECK_LESS(0, 0));
    check_no_throw(ARMARX_CHECK_LESS(-1, 1));
}

BOOST_AUTO_TEST_CASE(test_check_less_equal)
{
    check_no_throw(ARMARX_CHECK_LESS_EQUAL(1, 1));
    check_do_throw(ARMARX_CHECK_LESS_EQUAL(1, -1));
    check_no_throw(ARMARX_CHECK_LESS_EQUAL(0, 0));
    check_no_throw(ARMARX_CHECK_LESS_EQUAL(-1, 1));
}

BOOST_AUTO_TEST_CASE(test_check_greater)
{
    check_do_throw(ARMARX_CHECK_GREATER(1, 1));
    check_no_throw(ARMARX_CHECK_GREATER(1, -1));
    check_do_throw(ARMARX_CHECK_GREATER(0, 0));
    check_do_throw(ARMARX_CHECK_GREATER(-1, 1));
}

BOOST_AUTO_TEST_CASE(test_check_greater_equal)
{
    check_no_throw(ARMARX_CHECK_GREATER_EQUAL(1, 1));
    check_no_throw(ARMARX_CHECK_GREATER_EQUAL(1, -1));
    check_no_throw(ARMARX_CHECK_GREATER_EQUAL(0, 0));
    check_do_throw(ARMARX_CHECK_GREATER_EQUAL(-1, 1));
}

BOOST_AUTO_TEST_CASE(test_check_equal)
{
    std::vector<int> vec;
    check_no_throw(ARMARX_CHECK_EQUAL(1, 1));
    check_do_throw(ARMARX_CHECK_EQUAL(1, -1));
    check_no_throw(ARMARX_CHECK_EQUAL(0, 0));
    check_no_throw(ARMARX_CHECK_EQUAL(0, vec.size()));
    check_no_throw(ARMARX_CHECK_EQUAL(0u, 0));
    check_no_throw(ARMARX_CHECK_EQUAL(0, 0l));
    check_no_throw(ARMARX_CHECK_EQUAL(0.0, 0));
    check_no_throw(ARMARX_CHECK_EQUAL(0, 0.f));
    check_do_throw(ARMARX_CHECK_EQUAL(-1, 1));
}

BOOST_AUTO_TEST_CASE(test_check_not_equal)
{
    check_do_throw(ARMARX_CHECK_NOT_EQUAL(1, 1));
    check_no_throw(ARMARX_CHECK_NOT_EQUAL(1, -1));
    check_do_throw(ARMARX_CHECK_NOT_EQUAL(0, 0));
    check_no_throw(ARMARX_CHECK_NOT_EQUAL(-1, 1));
}

//            ARMARX_CHECK_POSITIVE
//            ARMARX_CHECK_NONNEGATIVE
//            ARMARX_CHECK_FITS_SIZE
//            ARMARX_CHECK_FINITE
//            ARMARX_CHECK_IS_NULL
//            ARMARX_CHECK_NOT_NULL
//            ARMARX_CHECK_EMPTY
//            ARMARX_CHECK_NOT_EMPTY
//            ARMARX_CHECK_MULTIPLE_OF
//            ARMARX_CHECK_NOT_MULTIPLE_OF

#define check_thrown_message(substring, ...)                                                       \
    try                                                                                            \
    {                                                                                              \
        __VA_ARGS__;                                                                               \
    }                                                                                              \
    catch (const armarx::LocalException& e)                                                        \
    {                                                                                              \
        BOOST_TEST_INFO(e.what());                                                                 \
        BOOST_CHECK(simox::alg::contains(e.what(), substring));                                    \
    }

BOOST_AUTO_TEST_CASE(test_message_check_condition)
{
    int a = 1, b = 3;
    int myCondition = a == b;
    check_thrown_message("a == b", ARMARX_CHECK(a == b));
    check_thrown_message("myCondition", ARMARX_CHECK(myCondition));
}

BOOST_AUTO_TEST_CASE(test_message_check_pointer)
{
    int* myPointer = nullptr;
    check_thrown_message("myPointer", ARMARX_CHECK(myPointer));
}

BOOST_AUTO_TEST_CASE(test_message_check_optional)
{
    std::optional<int> myOptional;
    check_thrown_message("myOptional", ARMARX_CHECK(myOptional));
    check_thrown_message("myOptional.has_value()", ARMARX_CHECK(myOptional.has_value()));
}

BOOST_AUTO_TEST_CASE(test_message_less_and_less_equal)
{
    int a = 10, b = 20;
    check_thrown_message("b < a", ARMARX_CHECK_LESS(b, a));
    check_thrown_message("b <= a", ARMARX_CHECK_LESS_EQUAL(b, a));
}

BOOST_AUTO_TEST_CASE(test_message_greater_and_greater_equal)
{
    int a = 10, b = 20;
    check_thrown_message("a > b", ARMARX_CHECK_GREATER(a, b));
    check_thrown_message("a >= b", ARMARX_CHECK_GREATER_EQUAL(a, b));
}

BOOST_AUTO_TEST_CASE(test_message_equal)
{
    int a = 10, b = 20;
    check_thrown_message("a == b", ARMARX_CHECK_EQUAL(a, b));
}

BOOST_AUTO_TEST_CASE(test_message_not_equal)
{
    int a = 20, b = 20;
    check_thrown_message("a != b", ARMARX_CHECK_NOT_EQUAL(a, b));
}

BOOST_AUTO_TEST_CASE(test_message_positive)
{
    int number = 0;
    check_thrown_message("number > 0", ARMARX_CHECK_POSITIVE(number));
}

BOOST_AUTO_TEST_CASE(test_message_nonnegative)
{
    int number = -1;
    check_thrown_message("number >= 0", ARMARX_CHECK_NONNEGATIVE(number));
}

BOOST_AUTO_TEST_CASE(test_message_fits_size)
{
    int index = 5;
    std::vector<int> vec{1, 2, 3};
    check_thrown_message("index", ARMARX_CHECK_FITS_SIZE(index, vec.size()));
    check_thrown_message("vec.size()", ARMARX_CHECK_FITS_SIZE(index, vec.size()));
}

BOOST_AUTO_TEST_CASE(test_message_close)
{
    float a = 0.125, b = 0.1250000001, prec = 1e-6;
    check_thrown_message("a", ARMARX_CHECK_CLOSE(a, b, prec));
    check_thrown_message("b", ARMARX_CHECK_CLOSE(a, b, prec));
    check_thrown_message("prec", ARMARX_CHECK_CLOSE(a, b, prec));
}

BOOST_AUTO_TEST_CASE(test_message_finite)
{
    float zero = 0;
    float a = 1 / zero;
    check_thrown_message("a", ARMARX_CHECK_FINITE(a));
}

BOOST_AUTO_TEST_CASE(test_message_null)
{
    int i = 2;
    int* myPointer = &i;
    check_thrown_message("myPointer", ARMARX_CHECK_NULL(myPointer));
}

BOOST_AUTO_TEST_CASE(test_message_is_null)
{
    int i = 2;
    int* myPointer = &i;
    check_thrown_message("myPointer", ARMARX_CHECK_IS_NULL(myPointer));
}

BOOST_AUTO_TEST_CASE(test_message_not_null)
{
    int* myPointer = nullptr;
    check_thrown_message("myPointer", ARMARX_CHECK_NOT_NULL(myPointer));
}

BOOST_AUTO_TEST_CASE(test_message_empty)
{
    std::vector<int> vec{1, 2, 3};
    check_thrown_message("vec", ARMARX_CHECK_EMPTY(vec));
}

BOOST_AUTO_TEST_CASE(test_message_not_empty)
{
    std::vector<int> vec;
    check_thrown_message("vec", ARMARX_CHECK_NOT_EMPTY(vec));
}

BOOST_AUTO_TEST_CASE(test_message_multiple_of)
{
    int mult = 101, of = 10;
    check_thrown_message("mult", ARMARX_CHECK_MULTIPLE_OF(mult, of));
    check_thrown_message("of", ARMARX_CHECK_MULTIPLE_OF(mult, of));
}

BOOST_AUTO_TEST_CASE(test_message_not_multiple_of)
{
    int mult = 100, of = 10;
    check_thrown_message("mult", ARMARX_CHECK_NOT_MULTIPLE_OF(mult, of));
    check_thrown_message("of", ARMARX_CHECK_NOT_MULTIPLE_OF(mult, of));
}
