/*
* This file is part of ArmarX.
*
* Copyright (C) 2011-2016, High Performance Humanoid Technologies (H2T), Karlsruhe Institute of Technology (KIT), all rights reserved.
*
* ArmarX is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License version 2 as
* published by the Free Software Foundation.
*
* ArmarX is distributed in the hope that it will be useful, but
* WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program. If not, see <http://www.gnu.org/licenses/>.
*
* @package    ArmarXCore::core::Test
* @author     Nils Adermann (naderman at naderman dot de)
* @date       2010
* @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
*             GNU General Public License
*/

#include "IceTestHelper.h"

#include <fstream>
#include <istream>

// From <ArmarXCore/Test.h>, but without Boost stuff.
namespace armarx::test
{
    std::map<std::string, std::string>
    readCMakeVars()
    {
        const std::filesystem::path armarxCoreDir = []()
        {
            char* armarxCoreDirC = std::getenv("ArmarXCore_DIR");
            ARMARX_CHECK_NOT_NULL(armarxCoreDirC);
            return std::filesystem::path{armarxCoreDirC};
        }();
        std::string test_H_Dir = armarxCoreDir / "testing/ArmarXCore/cmakevars.cfg";

        std::map<std::string, std::string> result;

        std::ifstream f(test_H_Dir);
        std::string line;
        while (std::getline(f, line))
        {
            size_t pos = line.find_first_of("=");
            if (pos != std::string::npos)
            {
                result[line.substr(0, pos)] = line.substr(pos + 1);
            }
        }
        return result;
    }

    std::string
    getCmakeValue(const std::string& varName)
    {
        static std::map<std::string, std::string> cmakeVars = readCMakeVars();
        return cmakeVars[varName];
    }
} // namespace armarx::test

IceTestHelper::IceTestHelper(int registryPort, int port) : registryPort(registryPort), port(port)
{
    std::stringstream defaultLocator;
    defaultLocator << "--Ice.Default.Locator=IceGrid/Locator:tcp -p " << registryPort << " -h "
                   << registryHost;
    args.push_back(defaultLocator.str());
    args.push_back("--Ice.ThreadPool.Client.SizeMax=20");
    args.push_back("--Ice.ThreadPool.Server.SizeMax=20");
    args.push_back("--Ice.Default.EncodingVersion=1.0 ");
    //        args.push_back("--Ice.Default.CollocationOptimized=0 ");
    args.push_back("--Ice.MessageSizeMax=20000000");

    Ice::StringSeq tempargs = args;
    ic = Ice::initialize(tempargs);
    init();
}

void
IceTestHelper::init()
{
    grid = false;
}

IceTestHelper::~IceTestHelper()
{
    armarx::LogSender::SetSendLoggingActivated(false);
    if (grid)
    {
        stopIceGrid();
    }
    ic->destroy();
}

void
IceTestHelper::startEnvironment()
{
    startIceGrid();
    //            admin = new armarx::IceGridAdmin(ic, "TestIceGrid"); <-- is already in armarxManager, so not needed here? (Mirko W)
    startIceStorm();

    ARMARX_INFO << "Sleeping for 10 s ...";
    sleep(10);

    auto makeIceGridAdminCommand = [this](const std::string& command)
    {
        std::stringstream ss;
        ss << "icegridadmin --host " << registryHost << " --port " << registryPort << " -u x -p y"
           << " -e \"" << command << "\"";
        return ss.str();
    };

    exec(makeIceGridAdminCommand("application list"));
    exec(makeIceGridAdminCommand("application describe IceStorm"));

    exec(makeIceGridAdminCommand("object list"));
    exec(makeIceGridAdminCommand("object describe"));

    exec(makeIceGridAdminCommand("server list"));
    exec(makeIceGridAdminCommand("server describe IceStorm"));

    exec(makeIceGridAdminCommand("service list IceStorm"));
    exec(makeIceGridAdminCommand("service describe IceStorm IceStorm"));

    exec(makeIceGridAdminCommand("adapter list"));
    exec(makeIceGridAdminCommand("adapter endpoints IceStorm.TopicManager"));
    exec(makeIceGridAdminCommand("adapter endpoints IceStorm.Publish"));
}

void
IceTestHelper::startIceStorm()
{
    deployApplication(armarx::test::getCmakeValue("ArmarXCore_CONFIG_DIR") +
                      "/IceStorm.icegrid.xml");
}

void
IceTestHelper::startIceGrid()
{
    const auto ucfgDirAbs = armarx::remove_trailing_separator(
        armarx::test::getCmakeValue("ARMARX_CONFIG_DIR_ABSOLUTE"));
    if (!std::filesystem::exists(ucfgDirAbs))
    {
        std::filesystem::create_directories(ucfgDirAbs);
    }
    std::string generatedDefaultTemplatePath =
        armarx::test::getCmakeValue("ArmarXCore_CONFIG_DIR") + "/default.generated.cfg";
    std::string generatedDefaultPath = ucfgDirAbs / "default.generated.cfg";
    //  std::filesystem::copy_file is buggy, so use system() instead
    //        std::filesystem::copy_file(generatedDefaultTemplatePath, generatedDefaultPath, std::filesystem::copy_option::overwrite_if_exists);
    exec(std::string("cp " + generatedDefaultTemplatePath + "  " + generatedDefaultPath));

    std::string defaultTemplatePath =
        armarx::test::getCmakeValue("ArmarXCore_CONFIG_DIR") + "/default.cfg";
    std::string defaultPath = ucfgDirAbs / "default.cfg";
    if (!std::filesystem::exists(defaultPath))
    {
        //  std::filesystem::copy_file is buggy, so use system() instead
        //std::filesystem::copy_file(defaultTemplatePath, defaultPath);
        exec(std::string("cp " + defaultTemplatePath + "  " + defaultPath));
    }

    configDir = (std::filesystem::temp_directory_path() / "cfgdir_").string() +
                std::to_string(std::random_device{}());
    configDir += "/icegridtestnode";
    ARMARX_INFO_S << "IceGrid Config dir: " << configDir;
    exec(std::string("mkdir -p ") + configDir + "/log");
    exec(std::string("mkdir -p ") + configDir + "/registry");
    exec(std::string("mkdir -p ") + configDir + "/data");

    std::string registryPortStr = std::to_string(registryPort);

    try
    {
        IceGrid::RegistryPrx prx =
            IceGrid::RegistryPrx::checkedCast(ic->stringToProxy("IceGrid/Registry"));

        if (!prx->createAdminSession("user", "password")->getAdmin()->pingNode("NodeMain"))
        {
            throw std::exception();
        }
    }
    catch (std::exception& e)
    {
        std::cout << "exception: " << e.what() << std::endl;
        IceUtil::ThreadControl::sleep(IceUtil::Time::milliSeconds(300));

        std::string command =
            std::string("cd ") + configDir + ";icegridnode " + "--daemon " + "--Ice.Config=\"" +
            armarx::test::getCmakeValue("ARMARX_CONFIG_DIR_ABSOLUTE") + "/default.generated.cfg," +
            armarx::test::getCmakeValue("ARMARX_CONFIG_DIR_ABSOLUTE") + "/default.cfg," +
            armarx::test::getCmakeValue("ArmarXCore_CONFIG_DIR") + "/icegrid_registry.cfg," +
            armarx::test::getCmakeValue("ArmarXCore_CONFIG_DIR") + "/armarx-icegridnode.cfg\" " +
            "--IceGrid.Registry.DefaultTemplates=\"" +
            armarx::test::getCmakeValue("ArmarXCore_CONFIG_DIR") +
            "/icegrid_default_templates.xml\" " +
            "--Ice.Default.Locator=\"IceGrid/Locator:tcp -p " + registryPortStr + " -h " +
            registryHost + "\" " + "--IceGrid.Registry.Client.Endpoints=\"tcp -p " +
            registryPortStr + "\" " + "--nochdir " + "--Ice.ProgramName=\"NodeMain\" " +
            "--IceGrid.Node.Name=\"NodeMain\" ";

        exec(command);
    }

    grid = true;
}

void
IceTestHelper::stopIceGrid()
{
    ARMARX_INFO << "Stopping IceGrid";
    std::string registryPortStr = std::to_string(registryPort);

    std::string gridCommand = std::string("icegridadmin ") +
                              "--Ice.Default.Locator=\"IceGrid/Locator:tcp -p " + registryPortStr +
                              " -h " + registryHost + "\" " + "-u x -p y " +
                              "-e \"node shutdown NodeMain\"";
    exec(gridCommand);

    IceUtil::ThreadControl::sleep(IceUtil::Time::milliSeconds(300));

    //            exec("killall icegridnode -9");
    //            exec("killall icebox -9");
    //            exec("killall icegridadmin -9");
}

void
IceTestHelper::deployApplication(const std::string& xmlPath)
{
    if (!std::ifstream(xmlPath.c_str()))
    {
        throw std::runtime_error(std::string("File not found: ") + xmlPath);
    }

    std::string registryPortStr = std::to_string(registryPort);

    std::string command = std::string("icegridadmin ") +
                          "--Ice.Default.Locator=\"IceGrid/Locator:tcp -p " + registryPortStr +
                          " -h " + registryHost + "\" " + "--Ice.PrintStackTraces=1 " +
                          "-u x -p y " + "-e \"application add " + xmlPath + "\" ";
    exec(command);
}

Ice::ObjectPrx
IceTestHelper::registerAdapter(const std::string& name, Ice::ObjectPtr object)
{
    std::string endpoint("tcp -p ");
    endpoint += std::to_string(port++);

    endpoints[name] = endpoint;

    Ice::ObjectAdapterPtr adapter = ic->createObjectAdapterWithEndpoints(name, endpoint);
    Ice::ObjectPrx prx = adapter->add(object, Ice::stringToIdentity(name));
    adapter->activate();
    return prx;
}

Ice::CommunicatorPtr
IceTestHelper::getCommunicator()
{
    return ic;
}

Ice::StringSeq
IceTestHelper::getArgs() const
{
    return args;
}

int
IceTestHelper::exec(const std::string& command)
{
    ARMARX_INFO << "Executing command: " << command;

    int res = std::system(command.c_str());
    res = WEXITSTATUS(res);

    ARMARX_INFO << "Exit status: " << res;
    return res;
}
