/*
 * This file is part of ArmarX.
 *
 * Copyright (C) 2011-2016, High Performance Humanoid Technologies (H2T), Karlsruhe Institute of Technology (KIT), all rights reserved.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @package    ArmarX
 * @author     Mirko Waechter( mirko.waechter at kit dot edu)
 * @date       2016
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */
#include "CycleUtil.h"

#include "TimeUtil.h"

namespace armarx
{
    CycleUtil::CycleUtil(IceUtil::Time const& cycleDuration, bool forceSystemTime) :
        cycleDuration(cycleDuration), forceSystemTime(forceSystemTime), busyWaitShare(0.0F)
    {
        reset();
    }

    CycleUtil::CycleUtil(const std::int64_t& cycleDurationMs, bool forceSystemTime) :
        cycleDuration(IceUtil::Time::milliSeconds(cycleDurationMs)),
        forceSystemTime(forceSystemTime)
    {
        reset();
    }

    void
    CycleUtil::reset()
    {
        cycleCount = 0;
        startTime = lastCycleTime = now();
        cycleMaxDuration = IceUtil::Time::microSeconds(0);
        cycleMinDuration = IceUtil::Time::microSeconds(std::numeric_limits<std::int64_t>::max());
    }

    IceUtil::Time
    CycleUtil::waitForCycleDuration()
    {
        auto start = armarx::CycleUtil::now();
        const auto waitTime = IceUtil::Time::microSeconds(
            std::max<std::int64_t>(0, (lastCycleTime + cycleDuration - start).toMicroSeconds()));
        const double waitTimeDouble = waitTime.toMicroSecondsDouble() -
                                      (busyWaitShare * cycleDuration.toMicroSecondsDouble());
        if (busyWaitShare < 1.0)
        {
            TimeUtil::Sleep(IceUtil::Time::microSecondsDouble(waitTimeDouble));
        }
        while ((now() - lastCycleTime).toMilliSecondsDouble() <
               cycleDuration.toMilliSecondsDouble())
        {
        }
        update();
        return armarx::CycleUtil::now() - start;
    }

    IceUtil::Time
    CycleUtil::update()
    {
        auto now = armarx::CycleUtil::now();
        auto cycleTime = now - lastCycleTime;
        cycleCount++;
        if (cycleTime > cycleMaxDuration)
        {
            cycleMaxDuration = cycleTime;
        }
        if (cycleTime < cycleMinDuration)
        {
            cycleMinDuration = cycleTime;
        }
        lastCycleTime = now;
        return cycleTime;
    }

    IceUtil::Time
    CycleUtil::getStartTime() const
    {
        return startTime;
    }

    std::int64_t
    CycleUtil::getCycleCount() const
    {
        return cycleCount;
    }

    IceUtil::Time
    CycleUtil::getAverageDuration() const
    {
        if (cycleCount > 0)
        {
            return IceUtil::Time::microSeconds((lastCycleTime - startTime).toMicroSeconds() /
                                               cycleCount);
        }
        return now();
    }

    IceUtil::Time
    CycleUtil::getMinimumDuration() const
    {
        return cycleMinDuration;
    }

    IceUtil::Time
    CycleUtil::getMaximumDuration() const
    {
        return cycleMaxDuration;
    }

    float
    CycleUtil::getBusyWaitShare() const
    {
        return busyWaitShare;
    }

    void
    CycleUtil::setBusyWaitShare(float value)
    {
        busyWaitShare = value;
    }

    IceUtil::Time
    CycleUtil::now()
    {
        struct timespec ts;
        clock_gettime(CLOCK_MONOTONIC, &ts);
        return IceUtil::Time::microSeconds(ts.tv_sec * 1e6 + ts.tv_nsec / 1000);
    }

    IceUtil::Time
    CycleUtil::getLastCycleTime() const
    {
        return lastCycleTime;
    }
} // namespace armarx
