/*
 * This file is part of ArmarX.
 *
 * Copyright (C) 2011-2016, High Performance Humanoid Technologies (H2T), Karlsruhe Institute of Technology (KIT), all rights reserved.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @package
 * @author
 * @date
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */
#include "CppClass.h"

#include <set>

#include <boost/format.hpp>

#include <SimoxUtility/algorithm/string/string_tools.h>

#include <ArmarXCore/core/exceptions/local/ExpressionException.h>


using namespace armarx;

CppClass::CppClass(const std::vector<std::string>& namespaces,
                   const std::string& name,
                   const std::string& templates) :
    MetaClass(name)
{
    this->namespaces = namespaces;
    this->templates = templates;
}

void
CppClass::write(const MetaWriterPtr& writer)
{
    CppWriterPtr w = std::dynamic_pointer_cast<CppWriter>(writer);
    writeCpp(w);
}

void
CppClass::WriteCpp(const std::vector<CppClassPtr>& classes, const CppWriterPtr& writer)
{
    for (const auto& meta : classes)
    {
        meta->writeCpp(writer);
    }
}

void
CppClass::writeCpp(const CppWriterPtr& writer)
{

    if (includes.size() > 0)
    {
        for (const auto& incl : includes)
        {
            writer->header.lineCheckDuplicate(boost::str(boost::format("#include %s") % incl));
        }
        writer->header.line();
    }

    for (const auto& ns : namespaces)
    {
        writer->body.startBlock(std::string("namespace ") + ns);
    }

    if (!docString.empty())
    {
        const std::string preSpace = " ";
        writer->body.line("/**");
        std::vector<std::string> docLines = simox::alg::split(docString, "\n");
        std::vector<std::string> blockLines;
        writer->ConvertToTextWithMaxWidth(80, docLines, blockLines);

        for (auto& line : blockLines)
        {
            writer->body.line(preSpace + line);
        }

        writer->body.line(" */");
    }
    if (!templates.empty())
    {
        writer->body.line(templates);
    }

    writer->body.line(std::string("class ") + name);

    for (size_t i = 0; i < inherit.size(); i++)
    {
        writer->body.line(boost::str(boost::format("%s%s%s") % (i == 0 ? ": " : "  ") %
                                     inherit.at(i) % (i < inherit.size() - 1 ? "," : "")),
                          1);
    }

    writer->body.startBlock();

    if (innerEnums.size() > 0)
    {
        writer->body.line("public:", -1);
        writer->body.line();

        for (const auto& inner : innerEnums)
        {
            inner->writeCpp(writer);
            writer->body.line();
        }
    }

    if (innerClasses.size() > 0)
    {
        writer->body.line("protected:", -1);
        writer->body.line();

        for (const auto& inner : innerClasses)
        {
            inner->writeCpp(writer);
            writer->body.line();
        }
    }

    if (privateFields.size() > 0)
    {
        writer->body.line("private:", -1);
        writer->body.line();

        for (const auto& field : privateFields)
        {
            field->writeCpp(writer);
            writer->body.line();
        }
    }

    if (protectedFields.size() > 0)
    {
        writer->body.line("protected:", -1);
        writer->body.line();

        for (const auto& field : protectedFields)
        {
            field->writeCpp(writer);
            writer->body.line();
        }
    }

    if (publicFields.size() > 0)
    {
        writer->body.line("public:", -1);
        writer->body.line();

        for (const auto& field : publicFields)
        {
            field->writeCpp(writer);
            writer->body.line();
        }
    }

    if (ctors.size() > 0)
    {
        writer->body.line("public:", -1);
        writer->body.line();

        for (const auto& ctor : ctors)
        {
            ctor->writeCpp(writer);
            writer->body.line();
        }
    }

    if (publicMethods.size() > 0)
    {
        writer->body.line("public:", -1);
        writer->body.line();

        for (const auto& method : publicMethods)
        {
            method->writeCpp(writer);
            writer->body.line();
        }
    }

    writer->body.endBlock("; // class " + name); // end of class

    for (std::vector<std::string>::reverse_iterator it = namespaces.rbegin();
         it != namespaces.rend();
         it++)
    {
        writer->body.endBlockComment(std::string("namespace ") + *it);
    }
}

void
CppClass::addPublicMethod(CppMethodPtr method)
{
    ARMARX_CHECK(method);
    publicMethods.push_back(method);
}

CppMethodPtr
CppClass::addPublicMethod(const std::string& header,
                          const std::string& doc,
                          const bool enforceBlockGeneration)
{
    CppMethodPtr method(new CppMethod(header, doc, enforceBlockGeneration));
    publicMethods.push_back(method);
    return method;
}

void
CppClass::addCtor(const CppCtorPtr& ctor)
{
    ctors.push_back(ctor);
}

CppCtorPtr
CppClass::addCtor(const std::string& arguments)
{
    CppCtorPtr ctor(new CppCtor(boost::str(boost::format("%s(%s)") % name % arguments)));
    ctors.push_back(ctor);
    return ctor;
}

CppCtorPtr
CppClass::addCtor(const std::string& arguments, const CppBlockPtr& b /* may be null */)
{
    CppCtorPtr ctor(new CppCtor(boost::str(boost::format("%s(%s)") % name % arguments), b));
    ctors.push_back(ctor);
    return ctor;
}

void
CppClass::addPrivateField(const CppFieldPtr& field)
{
    ARMARX_CHECK(field);
    privateFields.push_back(field);
}

void
CppClass::addProtectedField(const CppFieldPtr& field)
{
    ARMARX_CHECK(field);
    protectedFields.push_back(field);
}

void
CppClass::addPublicField(const CppFieldPtr& field)
{
    ARMARX_CHECK(field);
    publicFields.push_back(field);
}

CppClassPtr
CppClass::addInnerClass(const std::string& name)
{
    CppClassPtr innerClass(new CppClass(std::vector<std::string>(), name));
    innerClasses.push_back(innerClass);
    return innerClass;
}

void
CppClass::addInnerClass(CppClassPtr inner)
{
    innerClasses.push_back(inner);
}

void
CppClass::addInnerEnum(CppEnumPtr inner)
{
    ARMARX_CHECK(inner);
    ARMARX_CHECK(inner->getNamespaces().empty());
    innerEnums.push_back(inner);
}

CppEnumPtr
CppClass::addInnerEnum(const std::string& name)
{
    CppEnumPtr innerEnum(new CppEnum({}, name));
    innerEnums.push_back(innerEnum);
    return innerEnum;
}

void
CppClass::addInherit(const std::string& inherit)
{
    this->inherit.push_back(inherit);
}

void
CppClass::addInclude(const std::string include)
{
    includes.push_back(include);
}

bool
CppClass::hasInclude(const std::string include)
{
    return (std::find(includes.begin(), includes.end(), include) != includes.end());
}

void
CppClass::setTemplates(const std::string& s)
{
    templates = s;
}
