/*
 * This file is part of ArmarX.
 *
 * Copyright (C) 2011-2016, High Performance Humanoid Technologies (H2T), Karlsruhe Institute of Technology (KIT), all rights reserved.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @package
 * @author
 * @date
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */
#include "CppEnum.h"

#include <boost/format.hpp>

#include <SimoxUtility/algorithm/string/string_tools.h>

namespace armarx
{

    CppEnumField::CppEnumField(const std::string& name) : name(name), value(-1), hasValue(false)
    {
    }

    CppEnumField::CppEnumField(const std::string& name, int value) :
        name(name), value(value), hasValue(true)
    {
    }

    void
    CppEnumField::writeCpp(const CppWriterPtr& writer, bool addComma)
    {
        if (hasValue)
        {
            writer->body.line(name + " = " + std::to_string(value) + (addComma ? "," : ""));
        }
        return writer->body.line(name + (addComma ? "," : ""));
    }

    CppEnum::CppEnum(const std::vector<std::string> namespaces, const std::string& name) :
        MetaEnum(name), namespaces(namespaces)
    {
    }

    CppEnum::CppEnum(const std::vector<std::string> namespaces,
                     const std::string& name,
                     const std::vector<CppEnumFieldPtr>& fields) :
        MetaEnum(name), namespaces(namespaces), fields(fields)
    {
    }

    void
    CppEnum::write(const MetaWriterPtr& writer)
    {
        CppWriterPtr w = std::dynamic_pointer_cast<CppWriter>(writer);
        writeCpp(w);
    }

    void
    CppEnum::addField(const CppEnumFieldPtr& f)
    {
        fields.push_back(f);
    }

    std::vector<std::string>
    CppEnum::getNamespaces() const
    {
        return namespaces;
    }

    void
    CppEnum::writeCpp(const CppWriterPtr& writer)
    {
        for (const auto& ns : namespaces)
        {
            writer->body.startBlock(std::string("namespace ") + ns);
        }

        if (!docString.empty())
        {
            std::string delimiters = "\n";
            std::vector<std::string> doclines = simox::alg::split(docString, delimiters);
            writer->body.line("/**");

            for (auto& line : doclines)
            {
                writer->body.line(" * " + line);
            }

            writer->body.line(" */");
        }

        writer->body.line("enum class " + name);
        writer->body.startBlock();
        for (unsigned int i = 0; i < fields.size(); ++i)
        {
            auto& f = fields[i];
            f->writeCpp(writer, (i != (fields.size() - 1)));
        }
        writer->body.endBlock("; // enum class " + name); // end of enum

        for (std::vector<std::string>::reverse_iterator it = namespaces.rbegin();
             it != namespaces.rend();
             it++)
        {
            writer->body.endBlockComment(std::string("namespace ") + *it);
        }
    }
} // namespace armarx
