/*
 * This file is part of ArmarX.
 *
 * Copyright (C) 2011-2016, High Performance Humanoid Technologies (H2T), Karlsruhe Institute of Technology (KIT), all rights reserved.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @package
 * @author
 * @date
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */
#include "MetaWriter.h"

using namespace armarx;

Lines::Lines(const std::string& name, const std::string& i, const LineType t) :
    name(), indentChars(i), lastLineType(t)
{
}

void
Lines::startBlock()
{
    this->line("{");
    indent++;
    lastLineType = Lines::LineType::StartBlock;
}

void
Lines::startBlock(const std::string& line)
{
    this->line(line);
    this->line("{");
    indent++;
    lastLineType = Lines::LineType::StartBlock;
}

void
Lines::endBlock()
{
    indent--;
    this->line("}");
    lastLineType = Lines::LineType::EndBlock;
}

void
Lines::endBlock(const std::string& line)
{
    indent--;
    this->line(std::string("}") + line);
    lastLineType = Lines::LineType::EndBlock;
}

void
Lines::endBlockComment(std::string comment)
{
    indent--;
    this->line(std::string("} // ") + comment);
    lastLineType = Lines::LineType::EndBlock;
}

void
Lines::line(const std::string& line, int indentDelta)
{
    if (line.empty())
    {
        this->line();
        return;
    }

    if (const auto& it = std::find(lines.begin(), lines.end(), line); it == lines.end())
    {
        lines.push_back(line);
    }

    std::vector<std::string> docLines = simox::alg::split(line, "\n", false);
    for (auto& l : docLines)
    {
        lineInternal(l, indent + indentDelta);
    }
}

void
Lines::lineCheckDuplicate(const std::string& line, int indentDelta)
{
    // Empty lines are ok
    if (line.empty())
    {
        this->line();
        return;
    }

    if (const auto& it = std::find(lines.begin(), lines.end(), line); it == lines.end())
    {
        lines.push_back(line);

        std::vector<std::string> docLines = simox::alg::split(line, "\n", false);
        for (auto& l : docLines)
        {
            lineInternal(l, indent + indentDelta);
        }
    }
}

void
Lines::enforceEmptyLine()
{
    lineInternal("", indent);
}

void
Lines::line()
{
    switch (getLastLineType())
    {
        case Lines::LineType::Normal:
        case Lines::LineType::EndBlock:
            enforceEmptyLine();
            break;

        case Lines::LineType::Empty:
        case Lines::LineType::StartBlock:
            // no empty line needed
            break;
    }
}

Lines::LineType
Lines::getLastLineType()
{
    return lastLineType;
}

void
Lines::lineInternal(const std::string& line, int indent)
{
    for (int i = 0; i < indent; i++)
    {
        ss << indentChars;
    }

    ss << line << std::endl;
    lastLineType = line.empty() ? Lines::LineType::Empty : Lines::LineType::Normal;
}

MetaWriter::MetaWriter(const std::string& doc) :
    header("header"), body("body"), footer("footer"), fileDoc(doc)
{
}

void
MetaWriter::ConvertToTextWithMaxWidth(unsigned int maxTextWidth,
                                      const std::vector<std::string>& docLines,
                                      std::vector<std::string>& linesOut)
{
    for (std::string line : docLines)
    {
        while (line.size() > maxTextWidth)
        {
            auto pos = line.rfind(" ", maxTextWidth);
            linesOut.push_back(line.substr(0, pos));
            line.erase(0, pos);
        }

        linesOut.push_back(line);
    }
}

std::string
MetaWriter::getString()
{
    std::stringstream ss;
    ss << header.ss.str();
    ss << body.ss.str();
    ss << footer.ss.str();
    return ss.str();
}
