/*
* This file is part of ArmarX.
*
* Copyright (C) 2011-2016, High Performance Humanoid Technologies (H2T), Karlsruhe Institute of Technology (KIT), all rights reserved.
*
* ArmarX is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License version 2 as
* published by the Free Software Foundation.
*
* ArmarX is distributed in the hope that it will be useful, but
* WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program. If not, see <http://www.gnu.org/licenses/>.
*
* @package    ArmarX::Core
* @author     Kai Welke (welke _at_ kit _dot_ edu)
* @date       2012
* @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
*             GNU General Public License
*/

#pragma once

#include <memory>
#include <string>

#include <IceUtil/Handle.h>

#include <ArmarXCore/core/system/ImportExport.h>
#include <ArmarXCore/interface/observers/ObserverInterface.h>
#include <ArmarXCore/interface/observers/VariantBase.h>
#include <ArmarXCore/observers/condition/TermImpl.h>
#include <ArmarXCore/observers/variant/DataFieldIdentifier.h>
#include <ArmarXCore/observers/variant/Variant.h>

namespace armarx
{
    class LiteralImpl;

    /**
     * Typedef of LiteralImplPtr as IceInternal::Handle<LiteralImpl> for convenience.
     */
    using LiteralImplPtr = IceInternal::Handle<LiteralImpl>;

    /**
     * @class LiteralImpl
     * @brief LiteralImpls are the basic elements for defining conditional expressions.
     * @ingroup Conditions
     *
     * LiteralImpl are leaves in the expression tree. Each LiteralImpl is associated to a ConditionCheck. The state of
     * the LiteralImpl corresponds to the state of the ConditionCheck.
     */
    class ARMARXCORE_IMPORT_EXPORT LiteralImpl :
        virtual public LiteralImplBase,
        virtual public TermImpl
    {
        friend class ConditionHandler;
        friend class Literal;
        template <class BaseClass, class VariantClass>
        friend class GenericFactory;

    protected:
        /**
        * Creates an empty LiteralImpl. Required for Ice ObjectFactory
        */
        LiteralImpl();

        ~LiteralImpl();

        /**
        * Creates a LiteralImpl using string as dataFieldIdentifier with multiple parameter. See also createParameterList().
        *
        * @param dataFieldIdentifierStr identifier of the datafield used by the condition in the form "observerName.channelName.datafieldName"
        * @param checkName name of the condition check
        * @param checkParameter parameter for the condition check
        */
        LiteralImpl(const std::string& dataFieldIdentifierStr,
                    const std::string& checkName,
                    const ParameterList& checkParameters);

        /**
        * Creates a LiteralImpl using the DataFieldIdentifier type with multiple parameter. See also createParameterList()
        *
        * @param dataFieldIdentifier identifier of the datafield used by the condition
        * @param checkName name of the condition check
        * @param checkParameters list of check parameters
        */
        LiteralImpl(const DataFieldIdentifier& dataFieldIdentifier,
                    const std::string& checkName,
                    const ParameterList& checkParameters);

        /**
        * Creates a LiteralImpl using the DataFieldIdentifier type with multiple parameter. See also createParameterList()
        *
        * @param dataFieldIdentifier pointer to identifier of the datafield used by the condition
        * @param checkName name of the condition check
        * @param checkParameters list of check parameters
        */
        LiteralImpl(const DataFieldIdentifierPtr& dataFieldIdentifier,
                    const std::string& checkName,
                    const ParameterList& checkParameters);

        LiteralImpl(const DatafieldRefBasePtr& dataFieldIdentifier,
                    const std::string& checkName,
                    const ParameterList& checkParameters);

        void createInstance();

    public:
        /**
        * Retrieve check configuration. Inherited from LiteralImplBase
        *
        * @return configuration of check
        */
        CheckConfiguration
        getCheckConfiguration(const Ice::Current& c = Ice::emptyCurrent) override;

        /**
        * Set value of LiteralImpl
        *
        * @param value of LiteralImpl
        */
        void setValue(bool value, const Ice::Current& c = Ice::emptyCurrent) override;
        void setValueAndData(bool value,
                             const ::armarx::DataFieldIdentifierBasePtr& id,
                             const ::armarx::VariantBasePtr& data,
                             const Ice::Current& c = Ice::emptyCurrent) override;

        /**
        * Reimplementation of the ice_clone method.
        *
        * @return clone of the object
        */
        Ice::ObjectPtr ice_clone() const override;

        /**
         * output to stream. pure virtual.
         *
         * @param stream
         */
        void output(std::ostream& out) const override;

        /**
        * stream operator for LiteralImpl
        */
        friend std::ostream&
        operator<<(std::ostream& stream, const LiteralImpl& rhs)
        {
            rhs.output(stream);

            return stream;
        }

        /**
        * stream operator for LiteralImplPtr
        */
        friend std::ostream&
        operator<<(std::ostream& stream, const LiteralImplPtr& rhs)
        {
            rhs->output(stream);

            return stream;
        }

    protected:
        void init(const std::string& dataFieldIdentifierStr,
                  const std::string& checkName,
                  const ParameterList& checkParameters);

    private:
        void installCheck(const Ice::ObjectAdapterPtr& adapter, const ObserverInterfacePrx& proxy);
        void removeCheck(const Ice::ObjectAdapterPtr& adapter, const ObserverInterfacePrx& proxy);

        struct Impl;
        std::unique_ptr<Impl> impl;
    };
} // namespace armarx
extern template class ::IceInternal::Handle<::armarx::LiteralImpl>;
